==> dialserver-0.3.1.tar.gz: dialserver-0.3.1/server.c <==
// server.c -- Socket server thingy idea
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//
// Accepts multiple connections, and does reports and all that
// It is an interactive query idea -- but the client does not poll
// the server, it rather "LISTENS" for an event

#include "common.h"
#include "server.h"
#include "util.h"
#include "log.h"
#include "conf.h"

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

// ----------------------------------------------------------------
// -- Socket based server functions
// ----------------------------------------------------------------

CLIENTBUFFER client[MAX_CLIENTBUFFER];
#define LISTENERS 2
int listener[LISTENERS];
char *theunixsocket;

// Attempt to avoid "bind: port already in use" by closing the port
// (it's futile - methinks I misunderstand it)
void srv_cleanup(void)
{
	int a;
	for (a=0; a<MAX_CLIENTBUFFER; a++) if (client[a].fd) {
		 close(client[a].fd);
		 client[a].fd = 0;	// Just in case?
	}
	for (a=0; a<LISTENERS; a++) if (listener[a]) {
		close(listener[a]);
		listener[a] = 0;	// Equally in case
	}
	if (theunixsocket) {
		unlink(theunixsocket);
		theunixsocket = NULL;
	}
}

// A server what accepts and handles multiple connections 
// The callback function receives a ping ("" from a NULL client)
// or a transaction at least once every polltime seconds
int srv_startserver(int port, 
	   char *unixsocket,
	   void (*callback)(CLIENTBUFFER *client, char *request),
	   int polltime)
{
	struct protoent *pe;
	union {
		struct sockaddr_in addr_in;
		struct sockaddr_un addr_un;
	} addr;
	int a;
	int avail;
	fd_set exceptions, readable, writable;
	int nfds;
	struct timeval timeout;
	int bytes;
	char *p;

	atexit(srv_cleanup);

	// Set up TCP listener
	pe = getprotobyname("tcp");
	if (!pe) { perror("ds: getprotobyname"); return 1; }
	listener[0] = socket(AF_INET,SOCK_STREAM,pe->p_proto);
	if (listener[0]==-1) { perror("ds: socket(0)"); return 1; }
	memset(&addr.addr_in,0,sizeof(addr.addr_in));
	addr.addr_in.sin_family = AF_INET;
	addr.addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.addr_in.sin_port = htons(port);

	bytes = 1;
	setsockopt(listener[0],pe->p_proto, SO_REUSEADDR, &bytes, sizeof(bytes));
	if (bind(listener[0], (struct sockaddr*)(&addr.addr_in), sizeof( addr.addr_in))) {
		perror("tcp bind"); return 1; }

	bytes = 1;
	setsockopt(listener[0],pe->p_proto, TCP_NODELAY, &bytes, sizeof(bytes));
	if (listen(listener[0],10)) {
		perror("tcp listen"); return 1; }

	// Set up UNIX listener
	listener[1] = socket(PF_UNIX,SOCK_STREAM,0);  // man 7 unix
	if (listener[1]==-1) { perror("ds: socket(1)"); return 1; }
	memset(&addr.addr_un,0,sizeof(addr.addr_un));
	addr.addr_un.sun_family = AF_UNIX;
	strncpy(addr.addr_un.sun_path, unixsocket, sizeof(addr.addr_un.sun_path));
	unlink(unixsocket); // Kill any old socket
	if (bind(listener[1], (struct sockaddr*)(&addr.addr_un), sizeof(addr.addr_un))) {
		perror("unix bind"); return 1; }
	theunixsocket = unixsocket; // Make sure it gets killed when we exit
	if (listen(listener[1],10)) {
		perror("unix listen"); return 1; }
	chmod(unixsocket, 00777); // allow anything to all

	for (a=0 ; a<2; a++) 
		if (fcntl(listener[a], F_SETFL, O_NONBLOCK)) {
			perror("ds: fcntl"); return 1; }

keeplistening:
	FD_ZERO(&readable);
	FD_ZERO(&writable);
	FD_ZERO(&exceptions);  // Methinks exceptions is only for reading out-of-band stuff

	FD_SET(listener[0], &readable);
	FD_SET(listener[1], &readable);
	avail=-1;
	/* Assemble list of ports to listen to */
	for (a=0; a<MAX_CLIENTBUFFER; a++) {
		if (client[a].fd) {
			FD_SET(client[a].fd, &readable);
			if (client[a].outb)
				FD_SET(client[a].fd, &writable);
		}
		else {
			avail = a;
		}
	}
	timeout.tv_sec = polltime;
	timeout.tv_usec = 0;

	nfds = select(FD_SETSIZE, &readable, &writable, &exceptions, &timeout);
	if (nfds==-1 && errno!=EINTR ) {
		// This could be trouble - what if we ignore it? :)
		};
	if (nfds==0) {
		// Just a timeout?  Then do poll
		callback( NULL, "");
	}
	/* Well?  Who wants to talk to us? */
	for (a=0; a<LISTENERS; a++) if (listener[a]) {
		if (avail != -1 && FD_ISSET(listener[a], &readable)) {
			bytes = a==0 ? sizeof(addr.addr_in) : sizeof(addr.addr_un);
			if ((client[avail].fd = accept(
					listener[a], 
					( (struct sockaddr*)&addr ), 
					&bytes )) ==-1 ) 
			{
				client[avail].fd = 0; // forget it
			}
			else {
				if (!client[avail].outbuffer) {
					client[avail].outbuffer = xmalloc(SOCK_OUTBUFFER);
				}
				if (a==0) {  
					long address;
					short port;
					address = ntohl(addr.addr_in.sin_addr.s_addr);
					port    = ntohs(addr.addr_in.sin_port);
					snprintf(client[avail].peer,sizeof(client[avail].peer),
						"%d.%d.%d.%d:%d",
						(int)((address >> 24) & 0xFF),
						(int)((address >> 16) & 0xFF),
						(int)((address >> 8) & 0xFF),
						(int)((address >> 0) & 0xFF),
						(int)(port));
					bytes = 1;
					// Set socket option for TCP listener only
					setsockopt(client[avail].fd,pe->p_proto, TCP_NODELAY, &bytes, sizeof(bytes));
					fcntl(client[avail].fd, F_SETFL, O_NONBLOCK);
					if (
						// Allow only private addresses, and local
						((address & 0xFFFF0000) != (192<<24 | 168<<16)) &&
						((address & 0xFFFF0000) != (172<<24 |  16<<16)) &&
						((address & 0xFF000000) != ( 10<<24)) &&
						((address & 0xFF000000) != (127<<24))
						) {
						dlr_event("Dropping connection from %s\n",
							client[avail].peer);
						close(client[avail].fd);
						client[avail].fd = 0;
					}
				}
				else {	// Unix socket 
					fcntl(client[avail].fd, F_SETFL, O_NONBLOCK);
					snprintf(client[avail].peer,sizeof(client[avail].peer),
						"unix");
				}
				avail = -1;  // Don't reuse this socket...
			}
		}
	}
	// Handle incoming connections
	for (a=0; a<MAX_CLIENTBUFFER; a++) if (client[a].fd) {
		if (FD_ISSET(client[a].fd,&readable)) {
			bytes = recv(client[a].fd,
				client[a].inbuffer+client[a].inb,
				SOCK_INBUFFER-1-client[a].inb,
				MSG_NOSIGNAL);
			if (bytes==-1 && errno==EAGAIN) { 
				// I think this means that they have reached the end
			}
			else if (bytes==-1 && errno==EINTR) { // Hey, big deal 
			}
			else if (bytes<=0) { // Can't read?  Forget the whole business then.
				char *preserve;
				close(client[a].fd);
				preserve = client[a].outbuffer;
				memset(&client[a],0,sizeof(client[a]));
				client[a].outbuffer = preserve;
				if (bytes) perror("ds: recv");
			} 
			else {
				client[a].inb += bytes;
				do {
					int c;
					// Act on data received followed by CR or LF- call callback function
					p=NULL;
					for (c=0; c<client[a].inb; c++) 
						if (client[a].inbuffer[c]=='\n' ) 
					{
						p=client[a].inbuffer+c;
						if (c && p[-1]=='\r') p[-1]=0;
						*p=0;
						callback( &client[a], client[a].inbuffer);
						client[a].inb -= (c+1);
						memmove(client[a].inbuffer, p+1, client[a].inb);
						break;
					}
				} while (p!=NULL);
			}
		}
		// Write as much of pending response as we can
		if (client[a].outb && FD_ISSET(client[a].fd,&writable)) {
			bytes = send(client[a].fd,
				client[a].outbuffer, 
				client[a].outb,
				MSG_NOSIGNAL);
			if (bytes==-1 && errno==EINTR) { // Hey, big deal 
			}
			else if (bytes<=0) { // Can't write?  Duh?  Close the connection.
				char *preserve;
				if (bytes) perror("ds: send");
				close(client[a].fd);
				preserve = client[a].outbuffer;
				memset(&client[a],0,sizeof(client[a]));
				client[a].outbuffer = preserve;
			}
			else {
				// Chop it from the output buffer
				client[a].outb -= bytes;
				memmove(client[a].outbuffer,client[a].outbuffer+bytes,
					client[a].outb);
			}
		}
	}
	goto keeplistening;
	return 0;
}

// Print a formatted reply to the socket
void srv_printf(CLIENTBUFFER *client, char *format, ...)
{
	va_list ap;
	int max;
	int added;

	if (!client) { 
		return ; 
	}
	va_start(ap, format);
	/* Try to print in the allocated space. */
	max = SOCK_OUTBUFFER - 1 - client->outb;
	if (max<0) max=0;
	added = vsnprintf (client->outbuffer+client->outb, max, format, ap);
	if (added>0) 
		client->outb+=added;
	else 
		client->disconnect=1;  // And we just hope they figure out
			// what went wrong
	va_end(ap);
}

// Print a single character to the outut buffer
void srv_putc(CLIENTBUFFER *client, char c)
{
	if (!client) 
		return;
	else if (client->outb < SOCK_OUTBUFFER-1) 
		client->outbuffer[client->outb++]=c;
	else 
		client->disconnect=1;
		// And we just hope ditto
}

// A server: echo
void srv_echodebug(CLIENTBUFFER *client, char *request )
{
	srv_printf(client,"%s said '%s'\r\n",
		client->peer, request);
	return;
}

==> dialserver-0.3.3.tar.gz: dialserver-0.3.3/server.c <==
// server.c -- Socket server thingy idea
//
// Copyright (c) 2001,2002 Andrew McGill and Leading Edge Business Solutions
//
// This file is part of dialserver.
//
// dialserver is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// dialserver is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// dialserver; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA
//
// Accepts multiple connections, and does reports and all that
// It is an interactive query idea -- but the client does not poll
// the server, it rather "LISTENS" for an event

#include "common.h"
#include "server.h"
#include "util.h"
#include "log.h"
#include "conf.h"

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

// ----------------------------------------------------------------
// -- Socket based server functions
// ----------------------------------------------------------------

CLIENTBUFFER client[MAX_CLIENTBUFFER];
#define LISTENERS 2
int listener[LISTENERS];
char *theunixsocket;

// Attempt to avoid "bind: port already in use" by closing the port
// (it's futile - methinks I misunderstand it)
void srv_cleanup(void)
{
	int a;
	for (a=0; a<MAX_CLIENTBUFFER; a++) if (client[a].fd) {
		 close(client[a].fd);
		 client[a].fd = 0;	// Just in case?
	}
	for (a=0; a<LISTENERS; a++) if (listener[a]) {
		close(listener[a]);
		listener[a] = 0;	// Equally in case
	}
	if (theunixsocket) {
		unlink(theunixsocket);
		theunixsocket = NULL;
	}
}

// A server what accepts and handles multiple connections 
// The callback function receives a ping ("" from a NULL client)
// or a transaction at least once every polltime seconds
int srv_startserver(int port, 
	   char *unixsocket,
	   void (*callback)(CLIENTBUFFER *client, char *request),
	   int polltime)
{
	struct protoent *pe;
	union {
		struct sockaddr_in addr_in;
		struct sockaddr_un addr_un;
	} addr;
	int a;
	int avail;
	fd_set exceptions, readable, writable;
	int nfds;
	struct timeval timeout;
	int bytes;
	char *p;

	atexit(srv_cleanup);

	// Set up TCP listener
	pe = getprotobyname("tcp");
	if (!pe) { perror("ds: getprotobyname"); return 1; }
	listener[0] = socket(AF_INET,SOCK_STREAM,pe->p_proto);
	if (listener[0]==-1) { perror("ds: socket(0)"); return 1; }
	memset(&addr.addr_in,0,sizeof(addr.addr_in));
	addr.addr_in.sin_family = AF_INET;
	addr.addr_in.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.addr_in.sin_port = htons(port);

	bytes = 1;
	setsockopt(listener[0],pe->p_proto, SO_REUSEADDR, &bytes, sizeof(bytes));
	if (bind(listener[0], (struct sockaddr*)(&addr.addr_in), sizeof( addr.addr_in))) {
		perror("tcp bind"); return 1; }

	bytes = 1;
	setsockopt(listener[0],pe->p_proto, TCP_NODELAY, &bytes, sizeof(bytes));
	if (listen(listener[0],10)) {
		perror("tcp listen"); return 1; }

	// Set up UNIX listener
	listener[1] = socket(PF_UNIX,SOCK_STREAM,0);  // man 7 unix
	if (listener[1]==-1) { perror("ds: socket(1)"); return 1; }
	memset(&addr.addr_un,0,sizeof(addr.addr_un));
	addr.addr_un.sun_family = AF_UNIX;
	strncpy(addr.addr_un.sun_path, unixsocket, sizeof(addr.addr_un.sun_path));
	unlink(unixsocket); // Kill any old socket
	if (bind(listener[1], (struct sockaddr*)(&addr.addr_un), sizeof(addr.addr_un))) {
		perror("unix bind"); return 1; }
	theunixsocket = unixsocket; // Make sure it gets killed when we exit
	if (listen(listener[1],10)) {
		perror("unix listen"); return 1; }
	chmod(unixsocket, 00777); // allow anything to all

	for (a=0 ; a<2; a++) 
		if (fcntl(listener[a], F_SETFL, O_NONBLOCK)) {
			perror("ds: fcntl"); return 1; }

keeplistening:
	FD_ZERO(&readable);
	FD_ZERO(&writable);
	FD_ZERO(&exceptions);  // Methinks exceptions is only for reading out-of-band stuff

	FD_SET(listener[0], &readable);
	FD_SET(listener[1], &readable);
	avail=-1;
	/* Assemble list of ports to listen to */
	for (a=0; a<MAX_CLIENTBUFFER; a++) {
		if (client[a].fd) {
			FD_SET(client[a].fd, &readable);
			if (client[a].outb)
				FD_SET(client[a].fd, &writable);
		}
		else {
			avail = a;
		}
	}
	timeout.tv_sec = polltime;
	timeout.tv_usec = 0;

	nfds = select(FD_SETSIZE, &readable, &writable, &exceptions, &timeout);
	if (nfds==-1 && errno!=EINTR ) {
		// This could be trouble - what if we ignore it? :)
		};
	if (nfds==0) {
		// Just a timeout?  Then do poll
		callback( NULL, "");
	}
	/* Well?  Who wants to talk to us? */
	for (a=0; a<LISTENERS; a++) if (listener[a]) {
		if (avail != -1 && FD_ISSET(listener[a], &readable)) {
			bytes = a==0 ? sizeof(addr.addr_in) : sizeof(addr.addr_un);
			if ((client[avail].fd = accept(
					listener[a], 
					( (struct sockaddr*)&addr ), 
					&bytes )) ==-1 ) 
			{
				client[avail].fd = 0; // forget it
			}
			else {
				if (!client[avail].outbuffer) {
					client[avail].outbuffer = xmalloc(SOCK_OUTBUFFER);
				}
				if (a==0) {  
					long address;
					short port;
					address = ntohl(addr.addr_in.sin_addr.s_addr);
					port    = ntohs(addr.addr_in.sin_port);
					snprintf(client[avail].peer,sizeof(client[avail].peer),
						"%d.%d.%d.%d:%d",
						(int)((address >> 24) & 0xFF),
						(int)((address >> 16) & 0xFF),
						(int)((address >> 8) & 0xFF),
						(int)((address >> 0) & 0xFF),
						(int)(port));
					bytes = 1;
					// Set socket option for TCP listener only
					setsockopt(client[avail].fd,pe->p_proto, TCP_NODELAY, &bytes, sizeof(bytes));
					fcntl(client[avail].fd, F_SETFL, O_NONBLOCK);
					if (
						// Allow only private addresses, and local
						((address & 0xFFFF0000) != (192<<24 | 168<<16)) &&
						((address & 0xFFFF0000) != (172<<24 |  16<<16)) &&
						((address & 0xFF000000) != ( 10<<24)) &&
						((address & 0xFF000000) != (127<<24))
						) {
						dlr_event("Dropping connection from %s\n",
							client[avail].peer);
						close(client[avail].fd);
						client[avail].fd = 0;
					}
				}
				else {	// Unix socket 
					fcntl(client[avail].fd, F_SETFL, O_NONBLOCK);
					snprintf(client[avail].peer,sizeof(client[avail].peer),
						"unix");
				}
				avail = -1;  // Don't reuse this socket...
			}
		}
	}
	// Handle incoming connections
	for (a=0; a<MAX_CLIENTBUFFER; a++) if (client[a].fd) {
		if (FD_ISSET(client[a].fd,&readable)) {
			bytes = recv(client[a].fd,
				client[a].inbuffer+client[a].inb,
				SOCK_INBUFFER-1-client[a].inb,
				MSG_NOSIGNAL);
			if (bytes==-1 && errno==EAGAIN) { 
				// I think this means that they have reached the end
			}
			else if (bytes==-1 && errno==EINTR) { // Hey, big deal 
			}
			else if (bytes<=0) { // Can't read?  Forget the whole business then.
				char *preserve;
				close(client[a].fd);
				preserve = client[a].outbuffer;
				memset(&client[a],0,sizeof(client[a]));
				client[a].outbuffer = preserve;
				if (bytes) perror("ds: recv");
			} 
			else {
				client[a].inb += bytes;
				do {
					int c;
					// Act on data received followed by CR or LF- call callback function
					p=NULL;
					for (c=0; c<client[a].inb; c++) 
						if (client[a].inbuffer[c]=='\n' ) 
					{
						p=client[a].inbuffer+c;
						if (c && p[-1]=='\r') p[-1]=0;
						*p=0;
						callback( &client[a], client[a].inbuffer);
						client[a].inb -= (c+1);
						memmove(client[a].inbuffer, p+1, client[a].inb);
						break;
					}
				} while (p!=NULL);
			}
		}
		// Write as much of pending response as we can
		if (client[a].outb && FD_ISSET(client[a].fd,&writable)) {
			bytes = send(client[a].fd,
				client[a].outbuffer, 
				client[a].outb,
				MSG_NOSIGNAL);
			if (bytes==-1 && errno==EINTR) { // Hey, big deal 
			}
			else if (bytes<=0) { // Can't write?  Duh?  Close the connection.
				char *preserve;
				if (bytes) perror("ds: send");
				close(client[a].fd);
				preserve = client[a].outbuffer;
				memset(&client[a],0,sizeof(client[a]));
				client[a].outbuffer = preserve;
			}
			else {
				// Chop it from the output buffer
				client[a].outb -= bytes;
				memmove(client[a].outbuffer,client[a].outbuffer+bytes, client[a].outb);
			}
		}
	}
	goto keeplistening;
	return 0;
}

// Print a formatted reply to the socket
void srv_printf(CLIENTBUFFER *client, char *format, ...)
{
	va_list ap;
	int max;
	int added;

	if (!client) { 
		return ; 
	}
	va_start(ap, format);
	/* Try to print in the allocated space. */
	max = SOCK_OUTBUFFER - 1 - client->outb;
	if (max>0) {
		// Strange here .. if max==0, then it glibly prints like vsprintf
		added = vsnprintf (client->outbuffer+client->outb, max, format, ap);
	}
	else {
		added=0;
	}
	if (added>0) 
		client->outb+=added;
	else 
		client->disconnect=1;  // And we just hope they figure out
			// what went wrong
	va_end(ap);
}

// Print a single character to the outut buffer
void srv_putc(CLIENTBUFFER *client, char c)
{
	if (!client) 
		return;
	else if (client->outb < SOCK_OUTBUFFER-1) 
		client->outbuffer[client->outb++]=c;
	else 
		client->disconnect=1;
		// And we just hope ditto
}

// A server: echo
void srv_echodebug(CLIENTBUFFER *client, char *request )
{
	srv_printf(client,"%s said '%s'\r\n",
		client->peer, request);
	return;
}
